#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Configuration module

from typing import List
import os

SearchPath: List[str] = [os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext")]


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytecode optimizer.
# The Optimizer will attempt to optimize the bytecode that was emitted by the compiler by the following strategies:
# - remove push/pop pairs for the same register.

from __future__ import annotations
import pyximport; pyximport.install()
from .cmacal_vm import MacalVm
from .compiler_scope import CompilerScope
from .macal_instructions import MacalInstructionList
from .bytecode_address import Address

class BytecodeOptimizer:
    def __init__(self, vm: MacalVm, scope: CompilerScope):
        self.vm = vm
        self.scope = scope
        self.result = MacalVm()
        self.instructions = MacalInstructionList()

    def fix_jumps_for_push_pop_removal(addr):
        pass

    def get_label(self, lbl):
        for label in self.scope.labels:
            if label.label == lbl:
                return label

    def process(self):
        addr = 9
        pairs = 0
        jmp = 0
        fix = 0
        while addr < self.vm.rip.value:
            start = addr
            opcode = self.vm.memory._memory[addr]
            op = self.instructions.instructions[opcode]
            #  9 = PUSH
            # 10 = POP
            if opcode == 9 and self.vm.memory._memory[addr + 2] == 10 and self.vm.memory._memory[addr + 1] == self.vm.memory._memory[addr + 3]:
                pairs += 1
            if op.Name == 'JMP':
                jmp += 1
                for jump in self.scope.jump_table:
                    if jump.address == addr:
                        lbl = self.get_label(jump.label)
                        print(f"jump: {addr} {jump.address} -> {jump.label} {lbl.address}, {lbl.name}")
                        if lbl.address == addr + 9:
                            print("---------------------------> FOUND one")
                            fix+=1
                        next = self.vm.memory._memory[lbl.address]
                        nxtop = self.instructions.instructions[next]
                        if nxtop.Name == "JMP":
                            print("----------> FOUND JUMP TO JUMP")
            print(op.Name, ', '.join(op.Operands))
            if op.Length == -1:
                str = self.vm.peek_string(Address(addr+1))
                addr += len(str)
                addr += 2 # opcode + zero terminator for the string.
            elif op.Name == 'CALLE':
                addr += 9 # op + param count.
                str = self.vm.peek_string(Address(addr))
                addr += len(str)
                addr += 1
                str = self.vm.peek_string(Address(addr))
                addr += len(str)
                addr += 1
            elif op.Length == -3:
                #self.vm.show_memory(addr, 32)
                addr += 2 # the op and the register
                type = self.vm.opcode_to_type_map[self.vm.memory._memory[addr]-128]
                addr += 1 # the type
                rip = Address(addr)
                value = self.vm.peek_value(rip, type)
                addr = rip.address
            else:
                addr += op.Length

        print(pairs)
        print(jmp)
        print(fix)
        import sys
        sys.exit(0)




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the expression node for the AST.

from __future__ import annotations
from typing import Any, Optional
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .macal_conversions import convertToValue

class AstNodeLiteralExpression(AstNodeExpression):
    def __init__(self, token: LexToken, value: Optional[Any] = None, value_type: AstNodetype = AstNodetype.NIL) -> AstNodeLiteralExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.LITERAL_EXPRESSION
        if value is None:
            self._value: Optional[Any] = token.value
            self._value_type: AstNodetype = token.token_type
        else:
            self._value: Optional[Any] = value
            self._value_type: AstNodetype = value_type
        if self._value_type == AstNodetype.BOOLEAN:
            self._value = self._value == "true" or self._value is True



    def __str__(self) -> str:
        return f"AstNodeExpression({self._token})"



    def __repr__(self) -> str:
        return self.__str__()



    @property
    def value(self) -> Any:
        if self._value is not None:
            return self._value
        return convertToValue(self._token)



    @property
    def value_type(self) -> AstNodetype:
        return self._value_type



    def tree(self, indent: str = '') -> None:
        print(f"{indent}LiteralExpression:")
        print(f"{indent}    {self.value} ({self._token.token_type})")



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the new Macal Interpreter, it will lex, parse, compile and interpret (execute).

import os
import argparse
import pyximport; pyximport.install()
from macal.cmacal_vm import MacalVm
from macal.__about__ import __version__, __author__, __author_email__, __license__
from macal.config import SearchPath
from macal.macal_lexer import Lexer
from macal.macal_parser import Parser
from macal.macal_compiler import MacalCompiler
from macal.macal_decompiler import MacalDecompiler

bytecode_extension  = '.mbc'
source_extension = '.mcl'

def SetupSearchPath():
    search_path = [os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext")]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)

def check_file(file) -> str:
    if file is None: return
    if not file.endswith(bytecode_extension) and not file.endswith(source_extension):
        if os.path.exists(f"{file}{source_extension}"):
            file = f"{file}{source_extension}"
        else:
            raise Exception("Invalid Macal source file.")
    return file

def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description='Macal interpreter')
    parser.add_argument('file', nargs='?', type=check_file, help='Macal file')
    parser.add_argument('-d', '--decompile', action='store_true', help='Execute decompiler')
    parser.add_argument('-i', '--interactive', action='store_true', help='Interactive mode')
    parser.add_argument('-n', '--noexec', action='store_true', help='Do not execute')
    parser.add_argument('-o', '--output', help='Output file')
    parser.add_argument('-p', '--profile', action='store_true', help='Profile mode')
    parser.add_argument('-r', '--reserved', help='a csv list of reserved variables', default='', required=False, type=str)
    parser.add_argument('-s', '--verbose', action='store_true', help='Verbose mode')
    parser.add_argument('-v', '--version', action='version', version=f'Macal 2 interpreter v{__version__}')
    return parser

def LoadSource(fileName) -> str:
    with open(fileName, 'r') as f:
        text = f.read()
    return str(text)

def Execute(filename: str, profile: bool = False) -> int:
    vm = MacalVm(filename)
    if profile:
        import cProfile
        import pstats
        with cProfile.Profile() as pr:
            vm.Execute()
        result = pstats.Stats(pr)
        result.sort_stats('cumulative')
        result.print_stats()
    else:
        vm.Execute()
    return vm.stack[0][1]



def FromExecutable(args, file: str) -> None:
    if args.decompile:
        MacalDecompiler(file).Decompile()
    if not args.noexec:
        ShowResult(Execute(file, args.profile))

def FromSource(args, file: str, output_filename: str) -> None:
    source = LoadSource(file)
    tokens = Lexer(source).tokenize()
    if args.verbose:
        PrintTokens(tokens)
    ast = Parser(tokens, args.file).parse()
    if args.verbose:
        PrintAst(ast)
    reserved = None
    if args.reserved != '':
        reserved = args.reserved.split(',')
    MacalCompiler(program = ast, output_path = output_filename, verbose = args.verbose, reserved_vars=reserved).compile()
    if args.decompile:
        MacalDecompiler(output_filename).Decompile()
    if not args.noexec:
        ShowResult(Execute(output_filename, args.profile))

def ShowResult(exitCode: int) -> None:
    if exitCode == 0:
        print("Process exits OK")
    else:
        print(f"Process exits with error: {exitCode}")
    exit(exitCode)

def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.interactive:
        print("Interactive mode is not available yet.")
        return
        from macal.macal_interactive import MacalInteractive
        MacalInteractive().Run()
        return
    if args.file is None:
        parser.print_help()
        return
    output_filename = str(args.output or args.file.replace(source_extension, bytecode_extension))
    if args.file.endswith(bytecode_extension):
        FromExecutable(args, args.file)
        return
    FromSource(args, args.file, output_filename)

def PrintBanner() -> None:
    print(f"Macal interpreter v{__version__}")
    print(f"Author:  {__author__}")
    print(f"Email:   {__author_email__}")
    print(f"License: {__license__}")

def PrintTokens(tokens: list) -> None:
    print("Tokens:")
    for token in tokens:
        print(token)
    print()

def PrintAst(ast) -> None:
    print("AST:")
    ast.tree()
    print()

if __name__ == "__main__":
    Run()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


from typing import List, Any
from .bytecode_register import BytecodeRegister

class BytecodeFlagsRegister(BytecodeRegister):
    def __init__(self, name: str, opcode: int) -> BytecodeRegister:
        super().__init__(name, opcode)
        self.carry: bool = False
        self.zero: bool = False
        self.sign: bool = False
        self.overflow: bool = False
        self.direction: bool = False
        self.interrupt: bool = False
        self.trap: bool = False
        self.supervisor: bool = False
        self.negative: bool = False
        self.index: bool = False
        self.parity: bool = False



    def __str__(self) -> str:
        return f"CF {self._get_bit(0)} PF {self._get_bit(2)} ZF {self._get_bit(6)} SF {self._get_bit(7)} TF {self._get_bit(8)} DF {self._get_bit(10)} OF {self._get_bit(11)} IDX {self._get_bit(32)}"



    def __repr__(self) -> str:
        return self.__str__()



    def _get_bit(self, bit: int) -> str:
        return '1' if (self.value & (1 << bit)) != 0 else '0'



    def _set_bit(self, bit: int, value: bool) -> None:
        if value:
            self.value |= (1 << bit)
        else:
            self.value &= ~(1 << bit)




    def get_value(self):
        self._set_bit(0, self.carry)
        self._set_bit(1, True) # bit 1 is always true in EFLAGS
        self._set_bit(2, self.parity)
        self._set_bit(6, self.zero)
        self._set_bit(7, self.sign)
        self._set_bit(8, self.trap)
        self._set_bit(10, self.direction)
        self._set_bit(11, self.overflow)
        self._set_bit(32, self.index)
        return self.value



    def clear(self) -> None:
        self.carry = False # bit 0 (status)
        self._set_bit(1, True) # bit 1 is always true in EFLAGS
        self.parity = False # bit 2 (status)
        # bit 3 is reserved
        #self._auxiliary_carry = False # bit 4 (status)
        # bit 5 is reserved
        self.zero = False # bit 6 (status)
        self.sign = False # bit 7 (status)
        self.trap = False # bit 8 (single step) (control)
        #self._interrupt = False # bit 9 (interrupt enable) (control)
        self.direction = False # bit 10 (direction) (control)
        self.overflow = False # bit 11 (overflow) (status)
        self.index = False # bit 32 (status) (not a real flag, but used as such for the time being)
        # for RFLAGS bits 32-63 are reserved and must be 0



    def reset(self) -> None:
        self.clear()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Compiler helper for Macal

# write instructions to memory
# NOTE: When the instruction has no parameters, you must emit the tuple as (opcode,)
# If you do not provide the comma, pickle will deserialize it as a single integer.
# And that will fail when you try to execute it.

from __future__ import annotations
import sys
from typing import Any
from .ast_nodetype import AstNodetype
from .bytecode_register import BytecodeRegister

class MacalInstructionEmitter:
    def __init__(self, memory, rip: BytecodeRegister, rsp: BytecodeRegister, rbp: BytecodeRegister) -> MacalInstructionEmitter:
        self.memory = memory
        self.rip: BytecodeRegister = rip
        self.rsp: BytecodeRegister = rsp
        self.rbp: BytecodeRegister = rbp
        self.do_raise: bool = True # change this to false for production.


    def error(self, message: str) -> None:
        msg = f"Compiler Error: {message}"
        if self.do_raise:
            raise Exception(msg)
        print(msg)
        sys.exit(1)

    def NOP(self) -> None:
        self.memory[self.rip.value] = (0,)
        self.rip.value += 1
    
    def JMP(self, addr: int) -> None:
        self.memory[self.rip.value] = (1, addr)
        self.rip.value += 1
    
    def JMPZ(self, addr: int) -> None:
        self.memory[self.rip.value] = (2, addr)
        self.rip.value += 1
    
    def JMPNZ(self, addr: int) -> None:
        self.memory[self.rip.value] = (3, addr)
        self.rip.value += 1
    
    def JMPLT(self, addr: int) -> None:
        self.memory[self.rip.value] = (4, addr)
        self.rip.value += 1
    
    def JMPGT(self, addr: int) -> None:
        self.memory[self.rip.value] = (5, addr)
        self.rip.value += 1
    
    def JMPLTE(self, addr: int) -> None:
        self.memory[self.rip.value] = (6, addr)
        self.rip.value += 1
    
    def JMPGTE(self, addr: int) -> None:
        self.memory[self.rip.value] = (7, addr)
        self.rip.value += 1
    
    def DJNZ(self, addr: int) -> None:
        self.memory[self.rip.value] = (8, addr)
        self.rip.value += 1
    
    def PUSH(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (9, reg.opcode)
        self.rip.value += 1
        self.rsp.value += 1
    
    def POP(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (10, reg.opcode)
        self.rip.value += 1
        self.rsp.value -= 1
        #print(self.memory[0:self.rip.value-1])
        #sys.exit(1)
    
    def MOVRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (11, reg1.opcode, reg2.opcode)
        self.rip.value += 1
        if reg1.name == 'RBP' and reg2.name == 'RSP':
            self.rbp.value = self.rsp.value
        elif reg1.name == 'RSP' and reg2.name == 'RBP':
            self.rsp.value = self.rbp.value
    
    def MOVRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (12, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def MOVMI(self, addr, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (13, addr, node_type, value)
        self.rip.value += 1
    
    def MOVRM(self, reg: BytecodeRegister, addr) -> None:
        self.memory[self.rip.value] = (14, reg.opcode, addr)
        self.rip.value += 1
    
    def MOVMR(self, addr, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (15, addr, reg.opcode)
        self.rip.value += 1
    
    def CMPRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (16, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def CMPRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (17, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def ADDRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (18, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def ADDRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (19, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def SUBRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (20, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def SUBRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (21, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def MULRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (22, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def MULRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (23, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def DIVRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (24, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def DIVRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (25, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def MODRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (26, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def MODRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (27, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def POWRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (28, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def POWRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (29, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def ANDRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (30, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def ANDRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (31, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def ORRI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (32, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def ORRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (33, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def XORI(self, reg: BytecodeRegister, node_type: AstNodetype, value: Any) -> None:
        self.memory[self.rip.value] = (34, reg.opcode, node_type, value)
        self.rip.value += 1
    
    def XOR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (35, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def NEG(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (36, reg.opcode)
        self.rip.value += 1
    
    def NOT(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (37, reg.opcode)
        self.rip.value += 1
    
    def SETZR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (38, reg.opcode)
        self.rip.value += 1
    
    def SETNZR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (39, reg.opcode)
        self.rip.value += 1
    
    def SETLR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (40, reg.opcode)
        self.rip.value += 1
    
    def SETGR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (41, reg.opcode)
        self.rip.value += 1
    
    def SETLER(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (42, reg.opcode)
        self.rip.value += 1
    
    def SETGER(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (43, reg.opcode)
        self.rip.value += 1
    
    def CALL(self, addr: int) -> None:
        self.memory[self.rip.value] = (44, addr)
        self.rip.value += 1
        self.rsp.value += 1 # reserve space for the return value.
    
    def CALLR(self, reg: BytecodeRegister) -> None:
        if reg.name == "RAX":
            self.error("Can't use RAX in CALL/CALLR, RAX is reset to push a placeholder for the return value on the stack.")
        self.memory[self.rip.value] = (45, reg.opcode)
        self.rip.value += 1
        self.rsp.value += 1 # reserve space for the return value.
    
    def CALLE(self, module: str, function: str, param_count: int) -> None:
        self.memory[self.rip.value] = (46, module, function, param_count)
        self.rip.value += 1
        self.rsp.value += 1 # reserve space for the return value.
    
    def RET(self) -> None:
        self.memory[self.rip.value] = (47,)
        self.rip.value += 1
    
    def RETZ(self) -> None:
        self.memory[self.rip.value] = (48,)
        self.rip.value += 1
    
    def RETNZ(self) -> None:
        self.memory[self.rip.value] = (49,)
        self.rip.value += 1
    
    def HALT(self) -> None:
        self.memory[self.rip.value] = (50,)
        self.rip.value += 1
    
    def INC(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (51, reg.opcode)
        self.rip.value += 1
    
    def DEC(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (52, reg.opcode)
        self.rip.value += 1
    
    def LEN(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (53, reg.opcode)
        self.rip.value += 1
    
    def LENR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (54, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def INKEYS(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (55, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def LOAD(self, reg: BytecodeRegister, offset: int) -> None:
        self.memory[self.rip.value] = (56, reg.opcode, offset)
        self.rip.value += 1
    
    def STOR(self, offset: int, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (57, offset, reg.opcode)
        self.rip.value += 1
    
    def LOADR(self, reg1: BytecodeRegister, reg2: BytecodeRegister, offset: int) -> None:
        self.memory[self.rip.value] = (58, reg1.opcode, reg2.opcode, offset)
        self.rip.value += 1
    
    def STORR(self, reg1: BytecodeRegister, offset: int, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (59, reg1.opcode, offset, reg2.opcode)
        self.rip.value += 1
    
    def STORIR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (60, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def CMPRTI(self, reg: BytecodeRegister, node_type: AstNodetype) -> None:
        self.memory[self.rip.value] = (61, reg.opcode, node_type)
        self.rip.value += 1
    
    def SETRTI(self, reg: BytecodeRegister, node_type: AstNodetype) -> None:
        self.memory[self.rip.value] = (62, reg.opcode, node_type)
        self.rip.value += 1
    
    def APPEND(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (63, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def ERROR(self, errmsg) -> None:
        self.memory[self.rip.value] = (64, errmsg)
        self.rip.value += 1
    
    def ERRORZ(self, errmsg) -> None:
        self.memory[self.rip.value] = (65, errmsg)
        self.rip.value += 1
    
    def ERRORNZ(self, errmsg) -> None:
        self.memory[self.rip.value] = (66, errmsg)
        self.rip.value += 1
    
    def PRNT(self) -> None:
        self.memory[self.rip.value] = (67,)
        self.rip.value += 1
    
    def INDEX(self) -> None:
        self.memory[self.rip.value] = (68,)
        self.rip.value += 1
    
    def FEINDEX(self) -> None:
        self.memory[self.rip.value] = (69,)
        self.rip.value += 1
    
    def INDEXR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (70, reg.opcode)
        self.rip.value += 1
    
    def HASFLDRR(self, reg1: BytecodeRegister, reg2: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (71, reg1.opcode, reg2.opcode)
        self.rip.value += 1
    
    def MOVRTOA(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (72, reg.opcode)
        self.rip.value += 1
    
    def TYPE(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (73, reg.opcode)
        self.rip.value += 1
    
    def STACK(self) -> None:
        self.memory[self.rip.value] = (74,)
        self.rip.value += 1
    
    def REGS(self) -> None:
        self.memory[self.rip.value] = (75,)
        self.rip.value += 1
    
    def PRNTR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (76, reg.opcode)
        self.rip.value += 1
    
    def DATASEG(self, metadata: dict) -> None:
        self.memory[self.rip.value] = (77, metadata)
        self.rip.value += 1
    
    def MERGE(self, reg1: BytecodeRegister, reg2: BytecodeRegister, reg3: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (78, reg1.opcode, reg2.opcode, reg3.opcode)
        self.rip.value += 1
    
    def JMPR(self, reg: BytecodeRegister) -> None:
        self.memory[self.rip.value] = (79, reg.opcode)
        self.rip.value += 1
    
    def RESERVE(self, size: int) -> None:
        self.memory[self.rip.value] = (80, size)
        self.rip.value += 1
    
    def RESERVEDS(self, metadata: dict) -> None:
        self.memory[self.rip.value] = (81, metadata)
        self.rip.value += 1
    
    def LOADIR(self, reg1: BytecodeRegister, reg2: BytecodeRegister):
        self.memory[self.rip.value] = (82, reg1.opcode, reg2.opcode)
        self.rip.value += 1

    def MOVATOR(self, reg: BytecodeRegister):
        self.memory[self.rip.value] = (83, reg.opcode)
        self.rip.value += 1

    def DEBUGSTR(self, string: str):
        self.memory[self.rip.value] = (84, string)
        self.rip.value += 1
    
    def MOVSR(self, reg: BytecodeRegister):
        self.memory[self.rip.value] = (85, reg.opcode)
        self.rip.value += 1


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Lex token definitions for Macal

from __future__ import annotations
from .ast_node import AstNode
from .ast_nodetype import AstNodetype

class LexToken(AstNode):
    def __init__(self, type, value, lineno, offset, lexpos) -> LexToken:
        super().__init__()
        self.token_type  : AstNodetype = type
        self.value : str = value
        self.lineno: int = lineno
        self.offset: int = offset
        self.lexpos: int = lexpos

    def __str__(self) -> str:
        s = self.value
        if self.token_type in (AstNodetype.STRING, AstNodetype.STRING_INTERPOLATION_STRING_PART, AstNodetype.STRING_INTERPOLATION_END, AstNodetype.WHITESPACE):
            s = self.value.strip()
        return f"LexToken({self.token_type}, {s}, {self.lineno}, {self.offset} {self.lexpos})"

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def FromToken(token: LexToken) -> LexToken:
        return LexToken(token.token_type, token.value, token.lineno, token.lexpos, token.lexpos)


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This module contains the conversion utilities for the Macal language.

from typing import Any, Optional

from .ast_nodetype import AstNodetype
from .lex_token import LexToken



def convertToHex(value: int, digits: int = 4, prefix: str = '') -> str:
    if isinstance(value, int):
        return f'{prefix}{value:0{digits}X}'
    return f'{value}'.strip()



def convertToValue(token: LexToken) -> Any:
    if token.token_type == AstNodetype.INTEGER:
        return int(token.value)
    elif token.token_type == AstNodetype.FLOAT:
        return float(token.value)
    elif (token.token_type == AstNodetype.STRING
            or token.token_type == AstNodetype.STRING_INTERPOLATION_STRING_PART
            or token.token_type == AstNodetype.STRING_INTERPOLATION_END):
        #return token.value
        return token.value[1:-1]
    elif token.token_type == AstNodetype.BOOLEAN:
        return token.value == 'true'
    elif token.token_type == AstNodetype.BOOLEAN_TRUE:
        return True
    elif token.token_type == AstNodetype.BOOLEAN_FALSE:
        return False
    elif token.token_type == AstNodetype.NIL:
        return None
    else:
        raise Exception(f"convertToValue(): Unknown type {token.token_type}")



def compareTypes(type1: AstNodetype, type2: AstNodetype) -> bool:
    if type1 == type2:
        return True
    # handle the oddball cases
    if type1 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type1 == AstNodetype.STRING_INTERPOLATION_END:
        type1 = AstNodetype.STRING
    if type2 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type2 == AstNodetype.STRING_INTERPOLATION_END:
        type2 = AstNodetype.STRING
    return type1 == type2



def typeFromValue(value: Any) -> AstNodetype:
    # BOOLEAN BEFORE INTEGER as BOOLEAN is a subclass of INTEGER
    if isinstance(value, bool):
        return AstNodetype.BOOLEAN
    elif isinstance(value, int):
        return AstNodetype.INTEGER
    elif isinstance(value, float):
        return AstNodetype.FLOAT
    elif isinstance(value, str):
        return AstNodetype.STRING
    elif value is None or value == 'nil' or value == AstNodetype.NIL:
        return AstNodetype.NIL
    elif isinstance(value, list):
        return AstNodetype.ARRAY
    elif isinstance(value, dict):
        return AstNodetype.RECORD
    else:
        raise Exception(f"typeFromValue(): Unknown type {type(value)}")



def convertToHexAddr(addr: int) -> str:
    return f'0x{addr:08X}'



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# Bytecode debugging functions for Macal

from .macal_conversions import convertToHex, convertToHexAddr
from .ast_nodetype import AstNodetype

def ShowFlags(flags) -> None:
    print('╔══════════════════════RFLAGS═══════════════════════╗')
    print(f"║     {flags}      ║")
    print('╚═══════════════════════════════════════════════════╝')
    print()

def ShowRegisters(rax, rbx, rcx, rdx, rsi, rdi, r8, r9, r10, r11, r12, r13, r14, r15, rip, rsp, rbp, flags) -> None:
    print('╔══════════════════════════RFLAGS═════════════════════════╗')
    print(f"║        {flags}         ║")
    print('╠═════╦════════════╦═════╦════════════╦═════╦═════════════╣')
    print(f"║ RIP ║ {convertToHexAddr(rip.value)} ║ RSP ║ {convertToHexAddr(rsp.value)} ║ RBP ║ {convertToHexAddr(rbp.value)}  ║")
    print('╠═════╬════════════╬═════╩════════════╩═════╩═════════════╣')
    for i in [rax, rbx, rcx, rdx, rsi, rdi, r8, r9, r10, r11, r12, r13, r14, r15]:
        type = i.value_type.name
        value = f'{i.value}'
        if type == 'INTEGER':
            value = f'{convertToHexAddr(i.value)}'
        if len(value) > 36:
            value = value[:23] + '...'
        print(f"║ {i.name:>3} ║ {type:<10} ║ {value:<36} ║")
    print('╚═════╩════════════╩══════════════════════════════════════╝')
    print()

def ShowStack(stack, rsp, rbp) -> None:
    n = 8
    if n > len(stack):
        n = len(stack)
    print('╔═══════╦════════════╦══════╦STACK═════╦═════╦══════════════════════════╗')
    print(f'║  RSP  ║ {convertToHexAddr(rsp.value)} ║ SIZE ║ {convertToHex(len(stack), 8)} ║ TOP ║ {convertToHex(n, 8)}                 ║')
    print(f'║  RBP  ║ {convertToHexAddr(rbp.value)} ║      ║          ║     ║                          ║')
    print('╠═══════╬════════════╬══════╩══════════╩═════╩══════════════════════════╣')
    i = rsp.value
    if len(stack) > 0:
        while True:
            if i < 0: break
            number = f'{i}'
            if i == rsp.value:
                entry = None
                type = '-'
                value = '-'
            else:
                entry = stack[i]
                type = entry[0].name
                value = f'{entry[1]}'
                if entry[0] == AstNodetype.FUNCTION:
                    value = f'{convertToHexAddr(entry[1])} (->fn)'
            if len(value) > 48:
                value = value[:45] + '...'
            rbp_ptr = '>' if i == rbp.value else ' '
            print(f"║ {rbp_ptr} {number:>3} ║  {type:<8}  ║ {value:<48} ║")
            i -= 1
    print('╚═══════╩════════════╩══════════════════════════════════════════════════╝')
    print()

def ShowMemory(memory, address: int = 0, size: int = 256) -> None:
    print("Memory:")
    for i in range(address, address + size, 16):
        print(f"    {convertToHexAddr(i)}: ", end='')
        for j in range(16):
            if i+j >= len(memory._memory):
                print("     "*(16-j), end='')
                break
            print(f"{convertToHex(memory._memory[i+j], 2, '0x')} ", end='')
        for j in range(16):
            if i+j >= len(memory._memory):
                print("     ", end='')
                break
            if memory._memory[i+j] >= 32 and memory._memory[i+j] <= 126:
                print(f"{chr(memory._memory[i+j])}", end='')
            else:
                print('.', end='')
        print()
        if i >= len(memory._memory):
            break
    print()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for a select field.

from __future__ import annotations

from .ast_node import AstNode
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from typing import Optional

class AstNodeSelectField(AstNode):
    def __init__(self, token: LexToken, alt: Optional[LexToken] = None) -> AstNodeSelectField:
        super().__init__()
        self.expr_type = AstNodetype.SELECT_FIELD
        self.token: LexToken  = token
        self.altfield: Optional[LexToken] = alt
        if alt is None:
            self.altfield = token
        # good optimization, rather do it here, than having to loop through a dozen of if statements to determine if
        # we need to change the name of the field or not.
        # we now just retrieve the field from the record using the fieldname property, and we write it to the result
        # using the altfieldname property, not caring if it's different or the same.

    @property
    def altfieldname(self) -> Optional[str]:
        if self.altfield is not None:
            return self.altfield.value

    @property
    def fieldname(self) -> str:
        return self.token.value


    def __str__(self) -> str:
        return f"SelectField({self.fieldname}{'' if self.altfield is None else f' AS {self.altfieldname}'})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}SelectField: {self.fieldname}{'' if self.altfield is None else f' AS {self.altfieldname}'}")





#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This ast node type is for the Type statement.

from __future__ import annotations
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_expression import AstNodeExpression


class AstNodeTypeStatement(AstNodeStatement):
    def __init__(self, token: LexToken, var: AstNodeExpression):
        super().__init__(token)
        self.expr_type = AstNodetype.TYPE_STATEMENT
        self.expr: AstNodeExpression = var

    def __str__(self):
        return f"Type({self.expr})"

    def __repr__(self):
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}Type({self.expr})")



# cython: language_level=3
#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the main virtual machine engine for Macal.
# The code isn't pretty, but it is fast.

from __future__ import annotations
import os
import pickle
from typing import Any, Optional, Dict, List
import importlib
import importlib.util
from collections import deque
from .bytecode_register import BytecodeRegister
from .bytecode_flags_register import BytecodeFlagsRegister
from .ast_nodetype import AstNodetype
from .python_module_info import ModuleInfo
from .config import SearchPath
from .macal_conversions import typeFromValue
from .bytecode_debugger import ShowRegisters, ShowStack
from .runtime_scope import RuntimeScope

MAGIC_BYTES: bytes = b'BMC'
DEFAULT_MEMORY_SIZE: int = 0x100000 # 1 megabyte
FILE_FORMAT_VERSION: int = 1

class MacalVm:
    def __init__(self, filename: Optional[str] = None) -> MacalVm:
        self.filename = os.path.basename(filename) if filename is not None else None
        self.do_raise = True
        self.rax = BytecodeRegister('RAX',  0) # RAX is the aritmetic register, it's used always.
        self.rbx = BytecodeRegister('RBX',  1) # RBX is used for binary expressions.
        self.rcx = BytecodeRegister('RCX',  2) # RCX is used for loops.
        self.rdx = BytecodeRegister('RDX',  3)
        self.rsi = BytecodeRegister('RSI',  4)
        self.rdi = BytecodeRegister('RDI',  5)
        self.rbp = BytecodeRegister('RBP',  6)
        self.rsp = BytecodeRegister('RSP',  7)
        self.r8  = BytecodeRegister('R8',   8)
        self.r9  = BytecodeRegister('R9',   9)
        self.r10 = BytecodeRegister('R10', 10)
        self.r11 = BytecodeRegister('R11', 11)
        self.r12 = BytecodeRegister('R12', 12)
        self.r13 = BytecodeRegister('R13', 13)
        self.r14 = BytecodeRegister('R14', 14)
        self.r15 = BytecodeRegister('R15', 15)
        self.rip = BytecodeRegister('RIP', 16)
        self.flags = BytecodeFlagsRegister('FLAGS', 17)
        self.free_registers: List[BytecodeRegister] = [self.rdx, self.rsi, self.rdi, self.r8, self.r9, self.r10, self.r11, self.r12, self.r13, self.r14, self.r15]
        self.stack: deque = deque([(AstNodetype.INTEGER, 0), (AstNodetype.INTEGER, 0)])
        self.rsp.value = 2
        self.rbp.value = 2
        # The order in which this appears is important!
        self.opcode_to_register_map: list = [self.rax, self.rbx, self.rcx, self.rdx, self.rsi, self.rdi, self.rbp, self.rsp, self.r8, self.r9, self.r10, self.r11, self.r12, self.r13, self.r14, self.r15, self.rip, self.flags]
        # The order in which this appears is important!
        self.opcode_to_type_map : List[AstNodetype] = [AstNodetype.INTEGER, AstNodetype.FLOAT, AstNodetype.STRING, AstNodetype.BOOLEAN, AstNodetype.ARRAY, AstNodetype.RECORD, AstNodetype.VARIABLE, AstNodetype.FUNCTION,
                                   AstNodetype.TYPE, AstNodetype.NIL, AstNodetype.LIBRARY, AstNodetype.LABEL, AstNodetype.BC_METADATA, AstNodetype.STRING_INTERPOLATION_STRING_PART, AstNodetype.STRING_INTERPOLATION_END]

        self.memory : list = [None] * DEFAULT_MEMORY_SIZE
        self.Halted: bool = False
        self.Exitcode: int = 0
        self.LoadedModules: Dict[str, ModuleInfo] = {}
        self.scope: RuntimeScope = RuntimeScope()
        if filename is not None:
            self.Load(filename, 0)



    def Reset(self) -> None:
        self.free_registers = [self.rdx, self.rsi, self.rdi, self.r8, self.r9, self.r10, self.r11, self.r12, self.r13, self.r14, self.r15]
        self.stack = []
        self.memory = [None] * DEFAULT_MEMORY_SIZE
        self.Halted = False
        self.Exitcode = 0
        self.LoadedModules = {}
        self.rax.reset()
        self.rbx.reset()
        self.rcx.reset()
        self.rdx.reset()
        self.rsi.reset()
        self.rdi.reset()
        self.rbp.reset()
        self.rsp.reset()
        self.r8.reset()
        self.r9.reset()
        self.r10.reset()
        self.r11.reset()
        self.r12.reset()
        self.r13.reset()
        self.r14.reset()
        self.r15.reset()
        self.rip.reset()
        self.flags.reset()



    def Load(self, filename: str, offset: int) -> None:
        with open(filename, 'rb') as f:
            magic = f.read(3)
            if magic != MAGIC_BYTES:
                self._Error(f"File {filename} is not a valid bytecode file.")
            version = int.from_bytes(f.read(4), byteorder='little')
            if version != FILE_FORMAT_VERSION:
                self._Error(f"File {filename} is not a valid bytecode file.")
            memory_size = int.from_bytes(f.read(4), byteorder='little')
            if memory_size > DEFAULT_MEMORY_SIZE:
                self._Error(f"File {filename} is not a valid bytecode file.")
            self.memory = pickle.loads(f.read(memory_size))
            self.rip.value = offset
        self.LoadReservedVariables(self.scope)



    def SetReservedVariable(self, name: str, value: Any) -> None:
        self.scope.SetVariable(name, value)
        


    def LoadReservedVariables(self, scope: RuntimeScope) -> None:
        i = len(self.memory) - 1
        rv = i
        if self.memory[i][0] == 81:
            rv = i
        elif self.memory[i-1] == 81:
            rv = i-1
        else:
            return
        metadata = self.memory[rv][1]
        if "RESERVED" not in metadata:
            self._Error("No reserved variables found.")
        for var in metadata["RESERVED"]:
            scope.DefineVariable(var[0], var[1])



    def Save(self, filename: str, size: int) -> None:
        with open(filename, 'wb') as f:
            f.write(MAGIC_BYTES)
            f.write(FILE_FORMAT_VERSION.to_bytes(4, byteorder='little'))
            code = pickle.dumps(self.memory[:size])
            code_size = len(code)
            f.write(code_size.to_bytes(4, byteorder='little'))
            f.write(code)



    def __str__(self) -> str:
        return f"MacalVm()"



    def __repr__(self) -> str:
        return self.__str__()



    def _Error(self, message: str) -> None:
        msg = f"VM Error: {message}"
        if self.do_raise:
            raise Exception(msg)
        print(msg)
        exit(1)



    # helper functions
    def __SetFlags(self, res: Any, node_type: AstNodetype, lhs: Any, rhs: Any) -> None:
        if node_type == AstNodetype.INTEGER:
            self.flags.zero = res == 0
            self.flags.carry = res < 0
        else:
            self.flags.zero = lhs == rhs
            self.flags.carry = lhs < rhs



    def GetFreeRegister(self) -> Optional[BytecodeRegister]:
        if len(self.free_registers) == 0:
            return None
        return self.free_registers.pop()



    def ReleaseRegister(self, reg: BytecodeRegister) -> None:
        self.free_registers.append(reg)



    def Execute(self):
        while True:
            opCode = self.memory[self.rip.value]
            self.rip.value += 1
            if   opCode[0] == 9 : # PUSH
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.stack.append((reg.value_type, reg.value))
                self.rsp.value += 1
            elif opCode[0] == 10: # POP
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                while len(self.stack) > self.rsp.value: self.stack.pop()
                reg.value_type, reg.value = self.stack.pop()
                self.rsp.value -= 1
            elif opCode[0] == 12: # MOVRI
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value_type = opCode[2]
                reg.value = opCode[3]
            elif opCode[0] == 58: # LOADR reg, [reg + offset]
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                offset = opCode[3]
                reg1.value_type, reg1.value  = self.stack[reg2.value + offset]
            elif opCode[0] == 11: # MOVRR
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1.value_type, reg1.value = (reg2.value_type, reg2.value)
            elif opCode[0] == 17: # CMPRR
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                type1 = reg1.value_type
                type2 = reg2.value_type
                value1 = reg1.value
                value2 = reg2.value
                # fix values for NIL type
                if type1 == AstNodetype.NIL or (type1 != AstNodetype.STRING and value1 == 'nil'): 
                    type1 = AstNodetype.NIL
                    value1 = AstNodetype.NIL
                if type2 == AstNodetype.NIL or (type2 != AstNodetype.STRING and value2 == 'nil'): # fix values for NIL type
                    type2 = AstNodetype.NIL
                    value2 = AstNodetype.NIL
                if type1 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type1 == AstNodetype.STRING_INTERPOLATION_END:
                    type1 = AstNodetype.STRING
                if type2 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type2 == AstNodetype.STRING_INTERPOLATION_END:
                    type2 = AstNodetype.STRING
                self.flags.zero = False
                self.flags.carry = False
                if type1 == type2:
                    if type1 == AstNodetype.INTEGER:
                        res = value1 - value2
                        self.flags.zero = res == 0
                        self.flags.carry = res < 0
                    else:
                        self.flags.zero = value1 == value2                       
                        if type1 == AstNodetype.STRING and type2 == AstNodetype.STRING:
                            self.flags.carry = reg1.value < reg2.value                        
                elif type1 == AstNodetype.NIL or type2 == AstNodetype.NIL:
                    self.flags.zero = value1 == value2
                elif type1 == AstNodetype.BOOLEAN:
                    self.flags.zero = value2 != None and value2 != 0 and value2 != False and value2 != 'nil' and value2 != AstNodetype.NIL and value1
                elif type2 == AstNodetype.BOOLEAN:
                    self.flags.zero = value1 != None and value1 != 0 and value1 != False and value1 != 'nil' and value1 != AstNodetype.NIL and value2
            elif opCode[0] == 38: # SETZR reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.zero is True
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 47: # RET
                self.rip.value_type, self.rip.value = self.stack.pop()
                self.rsp.value -= 1
                if self.rip.value == 0:
                    self.Halted = True
                    return -2
            elif opCode[0] == 44: # CALL addr
                self.rax.value = 0
                self.rax.value_type = AstNodetype.INTEGER
                self.stack.append((self.rax.value_type, self.rax.value)) # push return value
                self.stack.append((self.rip.value_type, self.rip.value)) # push return address
                self.rsp.value += 2
                self.rip.value = opCode[1]
            elif opCode[0] == 59: # STORR offset, reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                offset = opCode[2]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[3]]
                self.stack[reg1.value + offset] = (reg2.value_type, reg2.value)
            elif opCode[0] ==  1: # JMP addr
                self.rip.value = opCode[1]
            elif opCode[0] == 33: # ORRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1.value = reg1.value is True or reg2.value is True
                self.flags.zero = reg1.value is True
            elif opCode[0] == 16: # CMPRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                type1 = reg.value_type
                type2 = opCode[2]
                value = opCode[3]
                if type1 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type1 == AstNodetype.STRING_INTERPOLATION_END:
                    type1 = AstNodetype.STRING
                if type2 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type2 == AstNodetype.STRING_INTERPOLATION_END:
                    type2 = AstNodetype.STRING
                if type1 == type2:
                    if (type1 == AstNodetype.INTEGER):
                        res = reg.value - value
                        self.flags.zero = res == 0
                        self.flags.carry = res < 0
                    else:
                        self.flags.zero = reg.value == value
                        self.flags.carry = reg.value < value
            elif opCode[0] ==  3: # JMPNZ addr
                if self.flags.zero is False:
                    self.rip.value = opCode[1]
            elif opCode[0] == 21: # SUBRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                if reg1.value_type == reg2.value_type:
                    if reg1.value_type == AstNodetype.INTEGER:
                        self.flags.zero = reg1.value - reg2.value == 0
                        self.flags.carry = reg1.value - reg2.value < 0
                    else:
                        self.flags.zero = reg1.value == reg2.value
                        self.flags.carry = reg1.value < reg2.value
                reg1 -= reg2
            elif opCode[0] == 19: # ADDRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                if reg1.value_type == reg2.value_type:
                    if reg1.value_type == AstNodetype.INTEGER:
                        self.flags.zero = reg1.value - reg2.value == 0
                        self.flags.carry = reg1.value - reg2.value < 0
                    else:
                        self.flags.zero = reg1.value == reg2.value
                        self.flags.carry = reg1.value < reg2.value
                reg1 += reg2
            elif opCode[0] == 46: # CALLE module, function, param_count
                # call external function.
                return_value  = self.run_external_function(opCode[1], opCode[2], opCode[3])
                self.rax.value = return_value
                self.rax.value_type  = typeFromValue(return_value)
            # The opcodes below are not used in the FIB test (function5) and are in random order.
            # Need to run a sensus in more applications to optimize the order.
            elif opCode[0] == 56: # LOAD reg, offset
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value_type, reg.value = self.stack[opCode[2]]
            elif opCode[0] == 57: # STOR offset, reg
                #print(opCode)
                #ShowStack(self.stack, self.rsp, self.rbp)
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.stack[opCode[1]] = (reg.value_type, reg.value)
            elif opCode[0] == 60: # STORIR reg1, reg2
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.stack[reg1.value] = (reg2.value_type, reg2.value)
            elif opCode[0] == 82: # LOADIR reg1, reg2
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1.value_type, reg1.value = self.stack[reg2.value]
            elif opCode[0] == 69: # FEINDEX
                if self.rax.value_type == AstNodetype.RECORD:
                    lng = len(self.rax.value.keys())
                    if self.rbx.value < 0 or self.rbx.value >= lng:
                        self._Error(f"FEINDEX: Record key index out of range: {self.rbx.value} (0..{lng-1}) (RIP: {self.rip.value-1})")
                    self.rax.value = list(self.rax.value.keys())[self.rbx.value]
                else:
                    if self.rbx.value_type != AstNodetype.INTEGER:
                        self._Error(f"FEINDEX: Type error, array index must be an integer. (RIP: {self.rip.value-1})")
                    if self.rbx.value < 0 or self.rbx.value >= len(self.rax.value):
                        self._Error(f"FEINDEX: Array index out of range: {self.rbx.value} (0..{len(self.rax.value)-1}) (RIP: {self.rip.value-1})")
                    self.rax.value = self.rax.value[self.rbx.value]
                self.rax.value_type = typeFromValue(self.rax.value)
            elif opCode[0] == 68: # INDEX
                if self.rax.value_type == AstNodetype.RECORD:
                    if self.rbx.value not in self.rax.value.keys():
                        self._Error(f"INDEX: Key error: {self.rbx.value} (RIP: {self.rip.value-1})")
                    self.rax.value = self.rax.value[self.rbx.value]
                elif self.rbx.value_type != AstNodetype.INTEGER:
                    self._Error(f"INDEX: Type error, array index must be an integer. (RIP: {self.rip.value-1})")
                elif self.rbx.value < 0 or self.rbx.value >= len(self.rax.value):
                    self._Error(f"INDEX: Array index out of range: {self.rbx.value} (0..{len(self.rax.value)-1}) (RIP: {self.rip.value-1})")
                else:
                    self.rax.value = self.rax.value[self.rbx.value]
                self.rax.value_type = typeFromValue(self.rax.value)
            elif opCode[0] == 70: # INDEXR
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                if self.rax.value_type != AstNodetype.ARRAY and self.rax.value_type != AstNodetype.RECORD:
                    self._Error(f"INDEXR: Type error, can't assign to an immutable object. (RIP: {self.rip.value-1})")
                if self.rax.value_type == AstNodetype.RECORD:
                    self.rax.value[self.rbx.value] = reg.value
                else:
                    if self.rbx.value_type != AstNodetype.INTEGER:
                        self._Error(f"INDEXR: Type error, array index must be an integer. (RIP: {self.rip.value-1})")
                    if self.rbx.value < 0 or self.rbx.value >= len(self.rax.value):
                        self._Error(f"INDEXR: Index out of range: {self.rbx.value} (0..{len(self.rax.value)-1}) (RIP: {self.rip.value-1})")
                    self.rax.value[self.rbx.value] = reg.value
                self.rax.value_type = typeFromValue(self.rax.value)
            elif opCode[0] == 63: # APPEND reg1, reg2
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                if reg1.value_type != AstNodetype.ARRAY:
                    self._Error(f"APPEND: Type error, can't append to an immutable object. (RIP: {self.rip.value-1})")
                reg1.value.append(reg2.value)
            elif opCode[0] == 39: # SETNZR reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.zero is False
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 40: # SETLR reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.carry is True and self.flags.zero is False
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 41: # SETGR reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.carry is False and self.flags.zero is False
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 42: # SETLER reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.carry is True or self.flags.zero is True
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 43: # SETGER reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = self.flags.carry is False or self.flags.zero is True
                reg.value_type = AstNodetype.BOOLEAN
            elif opCode[0] == 35: # XOR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1 ^= reg2
                self.flags.zero = reg1.value == 0
            elif opCode[0] == 45: # CALLR reg
                reg = self.opcode_to_register_map[opCode[1]]
                self.rax.reset()
                self.stack.append((self.rax.value_type, self.rax.value)) # push return value
                self.stack.append((self.rip.value_type, self.rip.value)) # push return address
                self.rsp.value += 2
                self.rip.value = reg.value
            elif opCode[0] == 67: # PRNT
                value = f"{self.rax.value}"
                if isinstance(self.rax.value, dict):
                    vd = {k: 'nil' if v is None or v == AstNodetype.NIL else str(v).lower() if isinstance(v, bool) else v 
                        for k, v in self.rax.value.items()}
                    value = f"{vd}"
                elif isinstance(self.rax.value, list):
                    vl = ['nil' if v is None or v == AstNodetype.NIL else str(v).lower() if isinstance(v, bool) else v 
                        for v in self.rax.value]
                    value = f"{vl}"
                elif self.rax.value_type == AstNodetype.NIL or self.rax.value == AstNodetype.NIL or self.rax.value is None:
                    value = "nil"
                elif self.rax.value_type == AstNodetype.BOOLEAN or isinstance(self.rax.value, bool):
                    value = "true" if self.rax.value == 1 or self.rax.value is True else "false"
                elif self.rax.value_type == AstNodetype.TYPE:
                    value = f"{self.opcode_to_type_map[self.rax.value-128].name.lower() }"
                print(f"{value}", end='')
            elif opCode[0] == 2: # JMPZ addr
                if self.flags.zero == True:
                    self.rip.value = opCode[1]
            elif opCode[0] == 4: # JMPLT addr
                if self.flags.zero == False and self.flags.carry == True:
                    self.rip.value = opCode[1]
            elif opCode[0] == 5: # JMPGT addr
                if self.flags.zero == False and self.flags.carry == False:
                    self.rip.value = opCode[1]
            elif opCode[0] == 6: # JMPLTE addr
                if (self.flags.zero == False and self.flags.carry == True) or (self.flags.zero == True and self.flags.carry == False):
                    self.rip.value = opCode[1]
            elif opCode[0] == 7: # JMPGTE addr
                if (self.flags.zero == False and self.flags.carry == False) or (self.flags.zero == True and self.flags.carry == False):
                    self.rip.value = opCode[1]
            elif opCode[0] == 13: # MOVMI [addr], type, value
                self.memory[opCode[1]] = (opCode[2], opCode[3])
            elif opCode[0] == 15: # MOVMR [addr], reg
                self.memory[opCode[1]] = (reg.value_type, reg.value)
            elif opCode[0] == 14: # MOVRM reg, [addr]
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value_type, reg.value = self.memory[opCode[2]]
            elif opCode[0] == 34: # XORI
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                tmp = BytecodeRegister('tmp', 0)
                tmp.value_type, tmp.value = (opCode[2], opCode[3])
                reg ^= tmp
                self.flags.zero = reg.value == 0
            elif opCode[0] == 48: # RETZ
                if self.flags.zero == True:
                    self.rip.value_type, self.rip.value = self.stack.pop()
                    self.rsp.value -= 1
                    if self.rip == 0:
                        self.Halted = True
                        return -2
            elif opCode[0] == 49: # RETNZ
                if self.flags.zero == False:
                    self.rip.value_type, self.rip.value = self.stack.pop()
                    self.rsp.value -= 1
                    if self.rip == 0:
                        self.Halted = True
                        return -2
            elif opCode[0] == 50: # HALT
                self.Halted = True
                return -2
            elif opCode[0] == 36: # NEG
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = -reg.value
            elif opCode[0] == 37: # NOT
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = not reg.value
            elif opCode[0] == 18: # ADDRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value_type, reg.value = (opCode[2], opCode[3])
            elif opCode[0] == 21: # SUBRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.__SetFlags(reg1.value - reg2.value, reg1.value_type, reg1.value, reg2.value)
                reg1 -= reg2
            elif opCode[0] == 20: # SUBRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.__SetFlags(reg.value - opCode[3], reg.value_type, reg.value, opCode[3])
                reg.value -= opCode[3]
            elif opCode[0] == 23: # MULRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.__SetFlags(reg1.value * reg2.value, reg1.value_type, reg1.value, reg2.value)
                reg1 *= reg2
            elif opCode[0] == 22: # MULRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.__SetFlags(reg.value * opCode[3], reg.value_type, reg.value, opCode[3])
                reg.value *= opCode[3]
            elif opCode[0] == 25: # DIVRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.__SetFlags(reg1.value / reg2.value, reg1.value_type, reg1.value, reg2.value)
                reg1 /= reg2
            elif opCode[0] == 24: # DIVRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.__SetFlags(reg.value / opCode[3], reg.value_type, reg.value, opCode[3])
                reg.value /= opCode[3]
            elif opCode[0] == 28: # POWRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.__SetFlags(reg1.value ** reg2.value, reg1.value_type, reg1.value, reg2.value)
                reg1 **= reg2
            elif opCode[0] == 29: # POWRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.__SetFlags(reg.value ** opCode[3], reg.value_type, reg.value, opCode[3])
                reg.value **= opCode[3]
            elif opCode[0] == 27: # MODRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.__SetFlags(reg1.value % reg2.value, reg1.value_type, reg1.value, reg2.value)
                reg1 %= reg2
            elif opCode[0] == 26: # MODRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.__SetFlags(reg.value % opCode[3], reg.value_type, reg.value, opCode[3])
                reg.value %= opCode[3]
            elif opCode[0] == 31: # ANDRR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1.value = reg1.value is True and reg2.value is True
                self.flags.zero = reg1.value is True
            elif opCode[0] == 30: # ANDRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = reg.value is True and opCode[3] is True
            elif opCode[0] == 32: # ORRI reg, type, value
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = reg.value is True or opCode[3] is True
                self.flags.zero = reg.value is True
            elif opCode[0] == 8: # DJNZ addr
                self.rcx.value -= 1
                self.flags.zero = self.rcx.value == 0
                if self.rcx.value != 0:
                    self.rip.value = opCode[1]
            elif opCode[0] == 61: # CMPRTI reg, type
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                self.flags.zero = reg.value_type == opCode[2]
            elif opCode[0] == 64: # ERROR msg
                self._Error(f"{opCode[1]}, (RIP: {self.rip.value-1})")
            elif opCode[0] == 65: # ERRORZ msg
                if self.flags.zero == True:
                    self._Error(f"{opCode[1]}, (RIP: {self.rip.value-1})")
            elif opCode[0] == 66: # ERRORNZ msg
                if self.flags.zero == False:
                    self._Error(f"{opCode[1]}, (RIP: {self.rip.value-1})")
            elif opCode[0] == 51: # INC reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value += 1
            elif opCode[0] == 52: # DEC reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value -= 1
            elif opCode[0] == 53: # LEN reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = len(reg.value)
                reg.value_type = AstNodetype.INTEGER
            elif opCode[0] == 54: # LENR reg, reg
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg1.value = len(reg2.value)
                reg1.value_type = AstNodetype.INTEGER
            elif opCode[0] == 62: # SETRTI reg, type
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value_type = opCode[2]
            elif opCode[0] == 72: # MOVRTOA
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = [reg.value]
                reg.value_type = AstNodetype.ARRAY
            elif opCode[0] == 83: # MOVATOR
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = reg.value[0]
                reg.value_type = typeFromValue(reg.value) # just in case it's a list of lists.
            elif opCode[0] == 71: # HASFLDRR record has field function
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                self.flags.zero = reg2.value in reg1.value.keys()
            elif opCode[0] == 73: # TYPE reg gets the type of the value of a register, basically puts reg.value_type in reg.value.
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg.value = reg.value_type
                reg.value_type = AstNodetype.TYPE
            elif opCode[0] == 0: # NOP
                pass
            elif opCode[0] == 74: # STACK
                ShowStack(self.stack, self.rsp, self.rbp)
            elif opCode[0] == 75: # REGS
                ShowRegisters(rax = self.rax, rbx = self.rbx, rcx = self.rcx, rdx = self.rdx,
                              rbp = self.rbp, rsp = self.rsp, rip = self.rip, rsi = self.rsi,
                              rdi = self.rdi,  r8 = self.r8,   r9 = self.r9,  r10 = self.r10,
                              r11 = self.r11, r12 = self.r12, r13 = self.r13, r14 = self.r14,
                              r15 = self.r15, flags = self.flags)
            elif opCode[0] == 76: # PRINTR reg
                reg: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                if reg.value_type == AstNodetype.NIL:
                    print(f"{reg.name}: nil (NIL)")
                value = f"{reg.value}"
                if reg.value_type == AstNodetype.BOOLEAN:
                    value = "true" if reg.value == 1 or reg.value is True else "false"
                print(f"{reg.name} {value} ({reg.value_type.name})")
            elif opCode[0] == 77: # DATASEG metadata
                pass
            elif opCode[0] == 78: # MERGE reg1, reg2, reg3 (this handles the merge for SELECT)
                reg1: BytecodeRegister = self.opcode_to_register_map[opCode[1]]
                reg2: BytecodeRegister = self.opcode_to_register_map[opCode[2]]
                reg3: BytecodeRegister = self.opcode_to_register_map[opCode[3]]
                if reg2.value_type == AstNodetype.RECORD:
                    reg2.value = [reg2.value]
                    reg2.value_type = AstNodetype.ARRAY
                if reg3.value_type == AstNodetype.RECORD:
                    reg3.value = [reg3.value]
                    reg3.value_type = AstNodetype.ARRAY
                if reg2.value_type != reg3.value_type != AstNodetype.ARRAY:
                    self._Error(f"MERGE: Type error, both parameters must be arrays. (RIP: {self.rip.value-1})")
                # just a single record in eachwith the same set of fields, then merge them.
                if len(reg2.value) == 0:
                    reg1.value = reg3.value.copy()
                elif len(reg3.value) == 0:
                    reg1.value = reg2.value.copy()
                elif len(reg2.value) == 1 and len(reg3.value) == 1 and set(reg2.value[0].keys()) == set(reg3.value[0].keys()):
                    reg1.value = [reg2.value[0], reg3.value[0]]
                # just a single record in each, but with different fields, then merge them but also the fields,
                # overwriting any existing in the source (reg2).
                elif len(reg2.value) == 1 and len(reg3.value) == 1:
                    keys = set().union(reg2.value[0].keys(), reg3.value[0].keys())
                    reg1.value = [{k: reg2.value[0].get(k, reg3.value[0].get(k, None)) for k in keys}]
                # multiple records in each, with the same set of fields, then just append them both.
                elif set(reg2.value[0].keys()) == set(reg3.value[0].keys()):
                    reg1.value = reg2.value.copy()
                    reg1.value.extend(reg3.value)
                # multiple records in each, but with the or different fields, then merge the records and fields.
                else:
                    keys = set().union(*(d.keys() for d in reg2.value + reg3.value))
                    reg1.value = [{k: d.get(k, None) for k in keys} for d in reg2.value + reg3.value]
                reg1.value_type = AstNodetype.ARRAY
                if len(reg1.value) == 1:
                    reg1.value = reg1.value[0]
                    reg1.value_type = AstNodetype.RECORD
            elif opCode[0] == 79: # JMP R
                self.rip.value = self.opcode_to_register_map[opCode[1]].value
            elif opCode[0] == 80: # RESERVE size entries on the stack.
                self.rsp.value += opCode[1]
                while len(self.stack) < self.rsp.value:
                    self.stack.append((AstNodetype.NIL, AstNodetype.NIL))
                for var in self.scope.variables:
                    if var.stack_offset < len(self.stack):
                        self.stack[var.stack_offset] = (var.node_type, var.value)
            elif opCode[0] == 81: # RESERVEDS metadata
                pass
            elif opCode[0] == 84: # DEBUGSTR string
                print(opCode[1])
            elif opCode[0] == 85: # MOVSR reg. get the top value from the stack without performing a pop.
                reg = self.opcode_to_register_map[opCode[1]]
                reg.value_type, reg.value = self.stack[self.rsp.value-1]
            else: break
            if self.flags.trap is True:
                self.Halted = True
                return -1



    def import_module(self, module_name: str) -> Optional[Any]:
        try:
            return importlib.import_module(module_name)
        except ImportError:
            return None



    def import_module_from_path(self, module_name: str) -> Optional[Any]:
        try:
            for path in SearchPath:
                path = os.path.join(path, f"{module_name}.py")
                if not os.path.exists(path): continue
                spec = importlib.util.spec_from_file_location(module_name, path)
                if spec is None: continue
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                return module
        except ImportError as ex:
            self._Error(f"Import error: {ex}")
        return None



    def run_external_function(self, module_name: str, function_name: str, param_count: int) -> Any:
        module = self.LoadedModules.get(module_name, None)
        if module is None:
            imported_module = self.import_module(module_name)
            if imported_module is None:
                imported_module = self.import_module_from_path(module_name)
                if imported_module is None:
                    self._Error(f"Module {module_name} not found.")
            module = ModuleInfo(module_name, imported_module)
            self.LoadedModules[module_name] = module
        function = module.functions.get(function_name, None)
        if function is None:
            self._Error(f"Function {function_name} not found in module {module_name}.")
        args = []
        # need to recalc stack_offset because it's possible for a params parameter to exist, and this can have many values.
        arg_count = self.stack[self.rbp.value - 4][1] # get the number of arguments from the stack.
        stack_offset = self.rbp.value - 4 - arg_count
        args = [self.stack[stack_offset+i][1] for i in range(arg_count)]
        return function(*args)
        



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# Macal math library external functions
#

from typing import Any
from math import floor, ceil, cos, acos, sin, asin, tan, atan, pow, sqrt, log, log2, log10, exp, expm1

def math_round(rval: Any, digits = None) -> Any:
    """Implementation of round function"""
    if rval is None:
        raise Exception('Round requires at least one argument.')
    if digits is None:
        return round(rval)
    else:
        return round(rval, digits)


def math_floor(rval) -> Any:
    """Implementation of floor function"""
    return floor(rval)

def math_ceil(rval: Any) -> Any:
    """Implementation of ceil function"""
    return ceil(rval)

def math_cos(rval: Any) -> Any:
    """Implementation of cos function"""
    return cos(rval)

def math_acos(rval: Any) -> Any:
    """Implementation of acos function"""
    return acos(rval)

def math_sin(rval: Any) -> Any:
    """Implementation of sin function"""
    return sin(rval)

def math_asin(rval: Any) -> Any:
    """Implementation of asin function"""
    return asin(rval)

def math_tan(rval: Any) -> Any:
    """Implementation of tan function"""
    return tan(rval)

def math_atan(rval: Any) -> Any:
    """Implementation of atan function"""
    return atan(rval)

def math_sqrt(rval: Any) -> Any:
    """Implementation of sqrt function"""
    return sqrt(rval)

def math_log(rval: Any) -> Any:
    """Implementation of log function"""
    return log(rval)

def math_log2(rval: Any) -> Any:
    """Implementation of log2 function"""
    return log2(rval)

def math_log10(rval: Any) -> Any:
    """Implementation of log10 function"""
    return log10(rval)

def math_exp(rval: Any) -> Any:
    """Implementation of exp function"""
    return exp(rval)

def math_expm1(rval: Any) -> Any:
    """Implementation of expm1 function"""
    return expm1(rval)



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# CSV Library external functions

from typing import Any
def headersToCsv(rec: Any) -> str:
    """Implementation of HeadersToCsv function"""
    separator = '","'
    return f'"{separator.join(rec)}"'

def valuesToCsv(rec: Any) -> str:
    """Implementation of ValuesToCsv function"""
    temp = []
    for fld in rec:
       temp.append(f'"{rec[fld]}"')
    separator = ','
    return separator.join(temp)

def arrayToCsv(arr: Any) -> str:
    """Implementation of ArrayToCsv function"""
    temp = []
    for fld in arr:
        temp.append(f'"{fld}"')
    separator = ','
    return separator.join(temp)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Lexer class that tokenizes the input string

from __future__ import annotations
from typing import Optional, List
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .statement_type_table import StatementTypeTable

class Lexer:
    def __init__(self, text: Optional[str] = None) -> Lexer:
        self.text: Optional[str] = text
        self.pos: int = 0
        if self.text is not None:
            self.current_char: Optional[str] = self.text[self.pos]
        self.lineno: int = 1
        self.offset: int = 1
        self.interpolate_start: bool = False
        self.interpolate_arg_count: int = 0
        self.interpolate_in_arg: bool = False
        self.interpolate_in_string: bool = False
        self.interpolate_terminator: Optional[str] = None
        self.interpolate_part_count: int = 0

    def error(self, msg) -> None:
        raise Exception(msg)

    def advance(self) -> None:
        self.pos += 1
        self.offset += 1
        if self.pos >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.pos]

    def peek(self) -> Optional[str]:
        peek_pos = self.pos + 1
        if peek_pos >= len(self.text):
            return None
        else:
            return self.text[peek_pos]

    def newline(self) -> None:
        if self.current_char == "\n":
            self.lineno += 1
            self.offset = 0

    def get_whitespace_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        while self.current_char is not None and self.current_char.isspace():
            self.newline()
            self.advance()
        return LexToken(AstNodetype.WHITESPACE, self.text[start:self.pos], line, offset, start)

    def get_number_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        type = AstNodetype.INTEGER
        while self.current_char is not None and self.current_char.isdigit():
            self.advance()
        if self.current_char == ".":
            type = AstNodetype.FLOAT
            self.advance()
            while self.current_char is not None and self.current_char.isdigit():
                self.advance()
        return LexToken(type,self.text[start:self.pos], line, offset, start)

    def do_esc(self, text: str) -> str:
        txt = text.replace("\\n", "\n").replace("\\t", "\t").replace("\\r", "\r")
        return txt

    def get_string_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        terminator = self.current_char
        self.advance()
        while self.current_char is not None and self.current_char != terminator:
            self.advance()
        self.advance()
        lexeme = self.do_esc(self.text[start:self.pos])
        return LexToken(AstNodetype.STRING, lexeme, line, offset, start)

    def get_identifier_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        while self.current_char is not None and (self.current_char.isalnum() or self.current_char == "_"):
            self.advance()
        identifier = self.text[start:self.pos]
        type = AstNodetype.IDENTIFIER
        if identifier in StatementTypeTable.keys():
            type = StatementTypeTable[identifier]
        return LexToken(type, identifier, line, offset, start)

    def get_short_comment_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        while self.current_char is not None and self.current_char != "\n":
            self.advance()
        return LexToken(AstNodetype.COMMENT, self.text[start:self.pos], line, offset, start)

    def strip_long_comment_text(self, text: str) -> str:
        return text.replace('\r', ' ').replace('\n', ' ').replace('\t', '    ')

    def get_long_comment_token(self) -> LexToken:
        start = self.pos
        line = self.lineno
        offset = self.offset
        self.advance() # skip /
        self.advance() # skip *
        while self.current_char is not None and self.current_char != "*" and self.peek() != "/":
            self.newline()
            self.advance()
        self.advance() # skip *
        self.advance() # skip /
        return LexToken(AstNodetype.COMMENT, self.strip_long_comment_text(self.text[start:self.pos]), line, offset, start)

    def get_string_interpolation_str_token(self, start, line, offset) -> LexToken:
        # this gets the string up until the next interpolation token.
        while (self.current_char is not None
                and self.current_char != self.interpolate_terminator
                and (self.current_char != "{" or
                        (self.current_char == "{" and self.peek() == "{"))):
            self.newline()
            self.advance()
        if self.current_char == self.interpolate_terminator:
            self.interpolate_in_string = False
            self.advance()
            value = self.text[start:self.pos]
            if not value.startswith(self.interpolate_terminator):
                value = self.interpolate_terminator + value
            return LexToken(AstNodetype.STRING_INTERPOLATION_END, value , line, offset, start)
        if self.current_char == "{": # next we expect expressions
            self.interpolate_in_arg = True
            self.advance()
            value = self.text[start:self.pos-1] # skip the {
            if not value.startswith(self.interpolate_terminator):
                value = self.interpolate_terminator + value
            if not value.endswith(self.interpolate_terminator):
                value = value + self.interpolate_terminator
            return LexToken(AstNodetype.STRING_INTERPOLATION_STRING_PART, value, line, offset, start)
        self.error(f"Invalid string interpolation at line {self.lineno} offset {self.offset}")

    def get_string_interpolation_start_token(self) -> LexToken:
        if self.current_char != "\"" and self.current_char != "'":
            self.error(f"Expected ' or \", got {self.current_char} at line {self.lineno} offset {self.offset}")
        self.interpolate_start = False
        self.interpolate_in_string = True
        self.interpolate_terminator = self.current_char
        start = self.pos
        line = self.lineno
        offset = self.offset
        self.advance() # skip the string terminator.
        return self.get_string_interpolation_str_token(start, line, offset)

    def get_next_token(self) -> LexToken:
        cur = self.current_char
        if cur is None:
            return LexToken(AstNodetype.EOF, "", self.lineno, self.offset, self.pos)
        if self.interpolate_start:
            return self.get_string_interpolation_start_token()
        if self.interpolate_in_string and not self.interpolate_in_arg:
            return self.get_string_interpolation_str_token(self.pos, self.lineno, self.offset)
        if self.interpolate_in_string and self.interpolate_in_arg and cur == '}':
            self.interpolate_in_arg = False
            self.interpolate_arg_count += 1
            self.advance()
            return self.get_string_interpolation_str_token(self.pos, self.lineno, self.offset)
        if cur.isspace():
            return self.get_whitespace_token()
        if cur.isdigit():
            return self.get_number_token()
        if cur == "\"" or self.current_char == "'":
            return self.get_string_token()
        if cur == "/" and self.peek() == "/":
            return self.get_short_comment_token()
        if cur == "#":
            return self.get_short_comment_token()
        if cur == "/" and self.peek() == "*":
            return self.get_long_comment_token()
        if cur.isalpha() or self.current_char == "_":
            return self.get_identifier_token()
        if cur == "+":
            self.advance()
            if self.current_char == "+":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_INCREMENT, "++", self.lineno, self.offset - 2, self.pos - 2)
            elif self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_INC, "+=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_ADDITION, "+", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "-":
            self.advance()
            if self.current_char == "-":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_DECREMENT, "--", self.lineno, self.offset - 2, self.pos - 2)
            elif self.current_char == '=':
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_DEC, "-=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_SUBTRACTION, "-", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "*":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_MUL, "*=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_MULTIPLICATION, "*", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "/":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_DIV, "/=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_DIVISION, "/", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "^":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_POW, "^=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_POWER, "^", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "%":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ASSIGNMENT_MOD, "%=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_MODULUS, "%", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "=":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.COMPARETOR_EQUAL, "==", self.lineno, self.offset - 2, self.pos - 2)
            if self.current_char == ">":
                self.advance()
                return LexToken(AstNodetype.OPERATOR_ARROW, "=>", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.OPERATOR_ASSIGNMENT, "=", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "<":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.COMPARETOR_LESS_THAN_EQUAL, "<=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.COMPARETOR_LESS_THAN, "<", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ">":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.COMPARETOR_GREATER_THAN_EQUAL, ">=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.COMPARETOR_GREATER_THAN, ">", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "!":
            self.advance()
            if self.current_char == "=":
                self.advance()
                return LexToken(AstNodetype.COMPARETOR_NOT_EQUAL, "!=", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.NOT_STATEMENT, "!", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "&":
            self.advance()
            if self.current_char == "&":
                self.advance()
                return LexToken(AstNodetype.AND_STATEMENT, "&&", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.BITWISE_AND, "&", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "|":
            self.advance()
            if self.current_char == "|":
                self.advance()
                return LexToken(AstNodetype.OR_STATEMENT, "||", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.BITWISE_OR, "|", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "~":
            self.advance()
            return LexToken(AstNodetype.BITWISE_NOT, "~", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "(":
            self.advance()
            return LexToken(AstNodetype.LEFT_PAREN, "(", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ")":
            self.advance()
            return LexToken(AstNodetype.RIGHT_PAREN, ")", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "[":
            self.advance()
            if self.current_char == "]":
                self.advance()
                return LexToken(AstNodetype.NEW_ARRAY, "[]", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.LEFT_BRACKET, "[", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "]":
            self.advance()
            return LexToken(AstNodetype.RIGHT_BRACKET, "]", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "{":
            self.advance()
            if self.current_char == "}":
                self.advance()
                return LexToken(AstNodetype.NEW_RECORD, "{}", self.lineno, self.offset - 2, self.pos - 2)
            return LexToken(AstNodetype.LEFT_BRACE, "{", self.lineno, self.offset - 1, self.pos - 1)
        if cur == "}":
            self.advance()
            return LexToken(AstNodetype.RIGHT_BRACE, "}", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ",":
            self.advance()
            return LexToken(AstNodetype.COMMA, ",", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ":":
            self.advance()
            return LexToken(AstNodetype.COLON, ":", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ";":
            self.advance()
            return LexToken(AstNodetype.SEMICOLON, ";", self.lineno, self.offset - 1, self.pos - 1)
        if cur == ".":
            self.advance()
            return LexToken(AstNodetype.DOT, ".", self.lineno, self.offset - 1, self.pos - 1)
        if cur == '$':
            self.advance()
            self.interpolate_start = True
            return LexToken(AstNodetype.STRING_INTERPOLATION_START, "$", self.lineno, self.offset - 1, self.pos - 1)
        self.error(f"Invalid character {self.current_char} at line {self.lineno} offset {self.offset}")

    def tokenize(self) -> List[LexToken]:
        if self.text is None:
            return []
        tokens = []
        token = self.get_next_token()
        while token.token_type != AstNodetype.EOF:
            if token is not None:
                tokens.append(token)
            token = self.get_next_token()
        tokens.append(token) # EOF
        return tokens
    
    def lex(self, text: str) -> List[LexToken]:
        if text is None:
            return []
        self.text = text
        self.pos = 0
        self.current_char = self.text[self.pos]
        self.lineno = 1
        self.offset = 1
        self.interpolate_start = False
        self.interpolate_arg_count = 0
        self.interpolate_in_arg = False
        self.interpolate_in_string = False
        self.interpolate_terminator = None
        self.interpolate_part_count = 0
        return self.tokenize()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytecode function for the compiler and the scope.

from __future__ import annotations

class BytecodeFunction:
    def __init__(self, name: str, memory_offset: int) -> BytecodeFunction:
        self.name: str = name
        self.memory_offset: int = memory_offset


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytecode label object

from __future__ import annotations
from .macal_conversions import convertToHexAddr

class BytecodeLabel:
    def __init__(self, label: int, address: int, name: str) -> BytecodeLabel:
        self.label: int = label
        self.address: int = address
        self.name: str = name

    def __str__(self) -> str:
        return f"BytecodeLabel(label={self.label}, address={convertToHexAddr(self.address)} name={self.name})"

    def __repr__(self) -> str:
        return self.__str__()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for the include statement.

from __future__ import annotations
from typing import List
from .ast_node_statement import AstNodeStatement
from .ast_node_library_expression import AstNodeLibraryExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeIncludeStatement(AstNodeStatement):
    def __init__(self, token: LexToken, libraries: List[AstNodeLibraryExpression]) -> AstNodeIncludeStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.INCLUDE_STATEMENT
        self.libraries: List[AstNodeLibraryExpression] = libraries

    def __str__(self) -> str:
        return f"IncludeStatement({self.libraries})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}IncludeStatement:")
        for lib in self.libraries:
            lib.tree(indent + '    ')


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for the halt statement.

from __future__ import annotations
from typing import Optional
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeHaltStatement(AstNodeStatement):
    def __init__(self, token: LexToken) -> AstNodeHaltStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.HALT_STATEMENT
        self.expr: Optional[AstNodeExpression] = None

    def __str__(self) -> str:
        return f"HaltStatement({self.expr})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}HaltStatement:")
        self.expr.tree(indent + '    ')



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


__version__ = '5.0.0'
__author__ = 'Marco Caspers'
__author_email__ = 'samadevteam@westcon.com'
__copyright__ = 'Copyright 2023 Westcon-Comstor'
__license__ = 'MIT License'


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This ast node type is for the IsXXXX type statements.

from __future__ import annotations
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_expression import AstNodeExpression


class AstNodeIsType(AstNodeStatement):
    def __init__(self, token: LexToken, var: AstNodeExpression, type: AstNodetype) -> AstNodeIsType:
        super().__init__(token)
        self.expr_type = AstNodetype.IS_TYPE_STATEMENT
        self.expr: AstNodeExpression = var
        self.TypeToCheck = type

    def __str__(self) -> str:
        return f"Is{self.TypeToCheck.name}({self.expr})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}Is{self.TypeToCheck.name}({self.expr})")



# Product:   Macal
# Author:    Marco Caspers
# Date:      23-11-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# Keyring Library external functions

import keyring

def GetPassword(username):
    return keyring.get_password("passwords", username)

def SetPassword(username, password):
    keyring.set_password("passwords", username, password)

def DeletePassword(username):
    keyring.delete_password("passwords", username)

def SetMerakiApiKey(customer, key):
    keyring.set_password("meraki", customer, key)

def GetMerakiApiKey(customer):
    return keyring.get_password("meraki", customer)


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is a unary expression AST node.

from __future__ import annotations
from .ast_nodetype import AstNodetype
from .ast_node_expression import AstNodeExpression
from .lex_token import LexToken

class AstNodeUnaryExpression(AstNodeExpression):
    def __init__(self, token: LexToken, right: AstNodeExpression) -> AstNodeUnaryExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.UNARY_EXPRESSION
        self.right: AstNodeExpression = right
        self.op: LexToken = token

    def __str__(self) -> str:
        return f"AstNodeUnaryExpression({self.token.value}, {self.right})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}UnaryExpression:")
        print(f"{indent}    {self.token.value} ({self.token.type}) {self.token.lineno}, {self.token.offset}, {self.token.lexpos}")
        self.right.tree(indent + '    ')



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the variable access node for the AST.

from __future__ import annotations
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeVariableExpression(AstNodeExpression):
    def __init__(self, variable: LexToken) -> AstNodeVariableExpression:
        super().__init__(variable)
        self.expr_type = AstNodetype.VARIABLE_EXPRESSION

    @property
    def name(self) -> str:
        return self._token.value

    @property
    def variable(self) -> LexToken:
        return self._token

    def __str__(self) -> str:
        return f"AstNodeVariableExpression({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = ''):
        print(f"{indent}AstNodeVariableExpression:")
        print(f"{indent}    {self.name} {self.line}, {self.column}")



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# IO Library external functions

import os
import json
from typing import Any



def loadTextFile(fileName: str) -> str:
    """Implementation of Load function"""    
    with open (fileName, "r") as tf:
        content=tf.read()
    return content



def readJSONFile(fileName: str) -> str:
    """Implementation of Read JSON file function"""
    with open(fileName, 'r') as fp:
        content = json.load(fp)
    return content



def existsFile(filename: str) -> bool:
    """Implementation of Exists function"""
    return os.path.exists(filename)



def saveTextFile(filename: str, content: str) -> bool:
    """Implementation of Save function"""
    with open(filename, "w") as tf:
        tf.write(content)
    return True



def writeJSONFile(filename: str, content: str) -> bool:
    """Implementation of Save function"""
    with open(filename, 'w') as fp:
        json.dump(content, fp, indent=4)
    return True



def getLastRun(org_name: str, default: Any) -> Any:
    """Implementation of GetLastRun function"""
    fileName = f"/tmp/last_run_{org_name}.ctl"
    if os.name == "nt":
        fileName = f"c:/temp/last_run_{org_name}.ctl"
    if os.path.exists(fileName):
        with open (fileName, "r") as tf:
            result=tf.read()
        if result is None or result == '':
            result = default
    else:
        result = default
    return result



def setLastRun(org_name: str, iso_now: Any) -> Any:
    """Implementation of SetLastRun function"""
    fileName = f"/tmp/last_run_{org_name}.ctl"
    if os.name == "nt":
        fileName = f"c:/temp/last_run_{org_name}.ctl"
    with open(fileName, "w") as tf:
        tf.write(iso_now)




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# function definition node for ast tree

from __future__ import annotations
from typing import List
from .ast_node import AstNode
from .ast_nodetype import AstNodetype
from .ast_node_block import AstNodeBlock
from .ast_node_function_parameter import AstNodeFunctionParameter
from .ast_node_expression import AstNodeExpression
from .ast_node_variable_expression import AstNodeVariableExpression
from .lex_token import LexToken

class AstNodeFunctionDefinition(AstNode):
    def __init__(self, var_expr: AstNodeVariableExpression, op: LexToken, parameters: List[AstNodeFunctionParameter], return_type: AstNodetype, body: AstNodeBlock) -> AstNodeFunctionDefinition:
        super().__init__()
        self.expr_type = AstNodetype.FUNCTION_DEFINITION
        self.return_type: AstNodetype = return_type
        self.var_expr: AstNodeVariableExpression = var_expr
        self.op: LexToken = op
        self.parameters : List[AstNodeFunctionParameter] = parameters
        self.body: AstNodeBlock = body
        self.external: bool = False
        self.module: str = ''
        self.function: str = ''

    @property
    def name(self) -> str:
        return self.var_expr.name

    @property
    def line(self) -> int:
        return self.var_expr.line

    @property
    def column(self) -> int:
        return self.var_expr.column

    @property
    def returns(self) -> bool:
        return self.return_type != AstNodetype.NIL

    def __str__(self) -> str:
        pl = [param.__str__() for param in self.parameters]
        return f"AstNodeFunctionDefinition({self.name}({', '.join(pl)})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}{self}")
        if self.external:
            print(f"{indent}    external: {self.module}.{self.function}")
        else:
            self.body.tree(indent + '    ')


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Compile time variable

from __future__ import annotations
from typing import Any, List
from .ast_nodetype import AstNodetype
from .ast_node_expression import AstNodeExpression

class CTVariable:
    # scope is a CompilerScope, but unfortunately we can't import it here because of circular imports.
    def __init__(self, name: str, scope) -> CTVariable:
        self.name: str = name
        self.node_type: AstNodetype = AstNodetype.NIL
        self.value: Any = None
        self.stack_offset: int = 0
        self.index: List[AstNodeExpression] = []
        self.scope = scope
        self.const = False

    def __str__(self) -> str:
        return f"CTVariable({self.name}, {self.node_type}, {self.value}, {self.stack_offset})"

    def __repr__(self) -> str:
        return self.__str__()





#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the node object for an record literal expression.

from __future__ import annotations
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeLiteralRecordExpression(AstNodeLiteralExpression):
    def __init__(self, token: LexToken, record: dict) -> AstNodeLiteralRecordExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.LITERAL_RECORD_EXPRESSION
        self.record: dict = record

    def tree(self, indent: str = '') -> None:
        print(f"{indent}LiteralRecordExpression:")
        print(f"{indent}  Record:")
        for key, value in self.record.items():
            print(f'    {indent}{key}: {value}')

    @property
    def value(self) -> dict:
        return self.record

    def __str__(self) -> str:
        return f"AstNodeLiteralRecordExpression({self.token})"

    def __repr__(self) -> str:
        return self.__str__()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the else statement AST node.

from __future__ import annotations
from .ast_nodetype import AstNodetype
from .ast_node_block import AstNodeBlock
from .lex_token import LexToken

class AstNodeElseStatement(AstNodeBlock):
    def __init__(self, token: LexToken) -> AstNodeElseStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.ELSE_STATEMENT

    def __str__(self) -> str:
        return f"AstNodeElseStatement({self.statements})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}ElseStatement:")
        super().tree(indent + '    ')




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the compile time function information.

from __future__ import annotations
from typing import List
from .macal_conversions import convertToHexAddr
from .ast_node_function_parameter import AstNodeFunctionParameter

class CTFunction:
    def __init__(self, start_address: int, name: str) -> CTFunction:
        self.start_address: int = start_address
        self.name: str = name
        self.return_type: str = ""
        self.external: bool = False
        self.module: str = ''
        self.function: str = ''
        self.parameters: List[AstNodeFunctionParameter] = []

    def __str__(self) -> str:
        return f"CTFunction({convertToHexAddr(self.start_address)}, {self.name})"

    def __repr__(self) -> str:
        return self.__str__()

class CTFunctionCall:
    def __init__(self, name: str, from_addr: int, to_addr) -> CTFunctionCall:
        self.name: str = name
        self.from_addr: int = from_addr
        self.to_addr: int = to_addr

    def __str__(self) -> str:
        return f"CTFunctionCall({self.name}, {convertToHexAddr(self.from_addr)} -> {convertToHexAddr(self.to_addr)})"

    def __repr__(self) -> str:
        return self.__str__()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytecode variable.

from __future__ import annotations
import sys
from typing import Any, List
from .ast_nodetype import AstNodetype
from .bytecode_register import BytecodeRegister

class BytecodeVariable:
    def __init__(self, name: str, type: AstNodetype, value: Any) -> BytecodeVariable:
        self.name: str = name
        if not isinstance(type, AstNodetype):
            raise TypeError(f"Expected AstNodetype, got {type}")
        self.type: AstNodetype = type
        self.index: List[Any] = []
        self.value: Any = value
        self.stack_offset: int = 0
        self.uninitialized: bool = True
        self.undefined: bool = type == AstNodetype.UNDEFINED

    def is_assigned(self) -> bool:
        return not self.uninitialized

    def from_register(self, register: BytecodeRegister) -> None:
        self.value = register.value
        self.type = register.value_type
        self.uninitialized = False

    def to_register(self, register: BytecodeRegister) -> None:
        register.value = self.value
        register.value_type = self.type
        self.uninitialized = False

    def get_indexed_value(self) -> Any:
        value = self.value
        for index in self.index:
            value = value[index]
        return value

    def set_indexed_value(self, value: BytecodeRegister) -> None:
        if len(self.index) == 0:
            self.value = value
        elif len(self.index) == 1:
            self.value[self.index[0]] = value
        else:
            v = self.value
            for i in range(len(self.index)-1):
                v = v[self.index[i]]
            v[self.index[-1]] = value

    def __str__(self) -> str:
        return f"{self.name} {self.type} {self.value}"

    def __repr__(self) -> str:
        return f"{self.name} {self.type} {self.value}"

    def error(self, message: str) -> None:
        print(f"Error: {message}")
        sys.exit(1)

    def __eq__(self, other: BytecodeVariable) -> bool:
        return self.type == other.type and self.value == other.value

    def __ne__(self, other: BytecodeVariable) -> bool:
        return self.type != other.type or self.value != other.value

    def __lt__(self, other: BytecodeVariable) -> bool:
        if self.type != other.type:
            self.error(f"Cannot compare {self.type} with {other.type}")
        return self.value < other.value

    def __le__(self, other: BytecodeVariable) -> bool:
        if self.type != other.type:
            self.error(f"Cannot compare {self.type} with {other.type}")
        return self.value <= other.value

    def __gt__(self, other: BytecodeVariable) -> bool:
        if self.type != other.type:
            self.error(f"Cannot compare {self.type} with {other.type}")
        return self.value > other.value

    def __ge__(self, other: BytecodeVariable) -> bool:
        if self.type != other.type:
            self.error(f"Cannot compare {self.type} with {other.type}")
        return self.value >= other.value

    def __add__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot add {self.type} with {other.type}")
        self.value += other.value
        return self

    def __sub__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot subtract {self.type} with {other.type}")
        self.value -= other.value
        return self

    def __mul__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot multiply {self.type} with {other.type}")
        self.value *= other.value
        return self

    def __truediv__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot divide {self.type} with {other.type}")
        if other.value == 0:
            self.error(f"Cannot divide by zero")
        self.value /= other.value
        return self

    def __mod__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot modulus {self.type} with {other.type}")
        self.value %= other.value
        return self

    def __pow__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot power {self.type} with {other.type}")
        self.value **= other.value
        return self

    def __and__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot and {self.type} with {other.type}")
        self.value &= other.value
        return self

    def __or__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot or {self.type} with {other.type}")
        self.value |= other.value
        return self

    def __xor__(self, other: BytecodeVariable) -> BytecodeVariable:
        if self.type != other.type:
            self.error(f"Cannot xor {self.type} with {other.type}")
        self.value ^= other.value
        return self

    def __invert__(self) -> BytecodeVariable:
        if self.type != AstNodetype.BOOLEAN:
            self.error(f"Cannot invert {self.type}")
        self.value = not self.value
        return self




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# Switch Case Table

from __future__ import annotations

class SwitchCaseTable:
    def __init__(self, address: int, labels: dict) -> SwitchCaseTable:
        self.address: int = address
        self.labels: dict = labels


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the root, program node for the AST.

from __future__ import annotations
from .ast_nodetype import AstNodetype
from .ast_node_block import AstNodeBlock
from .lex_token import LexToken

class AstNodeProgram(AstNodeBlock):
    def __init__(self, filename: str) -> AstNodeProgram:
        super().__init__(LexToken(type=AstNodetype.PROGRAM, lineno=1, offset=1, lexpos=0, value=filename))
        self.expr_type = AstNodetype.PROGRAM
        self.filename: str = filename

    def __str__(self) -> str:
        return f"AstNodeProgram({self.statements})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self) -> None:
        print(f"Program: {self.filename}")
        for statement in self.statements:
            statement.tree('    ')



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the information for a bytecode jump instruction.

from __future__ import annotations
from .macal_conversions import convertToHexAddr

class BytecodeJump:
    def __init__(self, address: int, label: int) -> BytecodeJump:
        self.address: int = address
        self.label: int = label

    def __str__(self) -> str:
        return f"BytecodeJump(address={convertToHexAddr(self.address)}, label={self.label})"

    def __repr__(self) -> str:
        return self.__str__()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the AST Node for the CASE statement

from __future__ import annotations
from typing import Optional
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_node_block import AstNodeBlock
from .macal_conversions import convertToValue

class AstNodeCaseStatement(AstNodeStatement):
    def __init__(self, token: LexToken, expr: AstNodeExpression) -> AstNodeCaseStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.CASE_STATEMENT
        self.expr: AstNodeExpression = expr
        self.block: Optional[AstNodeBlock] = None


    def case_value(self):
        return convertToValue(self.expr._token)


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for the library expression.

from __future__ import annotations
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeLibraryExpression(AstNodeExpression):
    def __init__(self, library: LexToken) -> AstNodeLibraryExpression:
        super().__init__(library)
        self.expr_type = AstNodetype.LIBRARY_EXPRESSION
        self.name: str = library.value

    def __str__(self) -> str:
        return f"LibraryExpression({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}LibraryExpression:")
        print(f"{indent}    {self.name}")



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Compiler scope is used to keep track of certain things that are needed during compilation.

from __future__ import annotations
from typing import Dict, List, Optional, Union
from .bytecode_label import BytecodeLabel
from .bytecode_jump import BytecodeJump
from .ct_function import CTFunction
from .ct_variable import CTVariable
from .switch_case_table import SwitchCaseTable

class FunctionAlias:
    def __init__(self, alias: str, name: str) -> FunctionAlias:
        self.alias: str = alias
        self.name: str = name

class CompilerScope:
    def __init__(self, name: str, parent: CompilerScope) -> CompilerScope:
        self.name: str = name
        self.parent: CompilerScope = parent
        # symbol table
        self.functions: List[CTFunction] = []
        self._libraries: Dict[str, int] = {}
        self.variables: List[CTVariable] = []
        self.function_aliasses: List[FunctionAlias] = []

        self.labels: List[BytecodeLabel] = []
        self.jump_table: List[BytecodeJump] = []
        self.switch_jump_tables: List[SwitchCaseTable] = []

        self.can_search_parent: bool = True
        self.children: List[CompilerScope] = []

        self.continue_label: Optional[int] = None
        self.break_label: Optional[int] = None
        self._main_label_name = 'main'
        self.exit_label: Optional[int] = None # set by compiler in function definition compilate.
        self.rbp_offset: int = 0 # set by compiler in scope creation.
        self.is_function_definition = False

    @property
    def root(self) -> CompilerScope:
        if self.parent is None:
            return self
        return self.parent.root

    @property
    def label_count(self) -> int:
        return len(self.root.labels)

    @property
    def variable_count(self) -> int:
        return len(self.variables)

    @property
    def function_count(self) -> int:
        return len(self.functions)

    @property
    def libraries(self) -> Dict[str, int]:
        return self.root._libraries

    @property
    def library_count(self) -> int:
        return len(self.root.libraries)

    @property
    def main_label_name(self) -> str:
        return self.root._main_label_name

    def new_child(self, name: str, base_pointer) -> CompilerScope:
        name = f'{self.name}.{name}_{len(self.children)}'
        child = CompilerScope(name, self)
        child.rbp_offset = base_pointer
        self.children.append(child)
        return child

    def remove_child(self, child: CompilerScope) -> None:
        self.children.remove(child)

    def get_new_label(self, name: Optional[str] = None) -> Optional[int]:
        id = len(self.root.labels)
        if name is None:
            name = f'label'
        lbl_name = f'{name}_{id}' # ensure unique label name
        if name == "main":
            self.root._main_label_name = lbl_name
        #if any(label.name == lbl_name for label in self.root.labels):
        #    return None
        self.root.labels.append(BytecodeLabel(id, -1, lbl_name))
        return id

    def get_label(self, name: str) -> Optional[int]:
        return next((label.id for label in self.root.labels if label.name == name), None)

    def set_label_address(self, id: int, address: int) -> bool:
        if not isinstance(address, int):
            raise TypeError(f"address must be an int, not {type(address)}")
        if id < 0 or id >= len(self.root.labels):
            return False
        self.root.labels[id].address = address
        return True

    def new_jump(self, address: int, label_id: int) -> bool:
        self.root.jump_table.append(BytecodeJump(address, label_id))
        return True

    def define_variable(self, name: str, offset: int) -> Optional[CTVariable]:
        var = CTVariable(name, self)
        var.stack_offset = offset
        self.variables.append(var)
        return var

    def get_variable(self, name: str) -> Optional[Union[CTVariable, CTFunction]]:
        var = next((var for var in self.variables if var.name == name), None)
        if var is not None:
            return var
        if self.parent is not None and self.can_search_parent:
            var = self.parent.get_variable(name)
            if var is not None:
                return var
        funcname = f"{self.name}.{name}"
        func = next((func for func in self.functions if func.name == funcname), None)
        return func

    def find_variable(self, name: str) -> Optional[CTVariable]:
        var = next((var for var in self.variables if var.name == name), None)
        if var is not None: return var
        if self.parent is not None and self.can_search_parent:
            return self.parent.find_variable(name)
        return None

    def define_function(self, name: str, function_start_address: int) -> Optional[CTFunction]:
        if any(func.name == name for func in self.functions):
            return None
        func = CTFunction(function_start_address, name)
        self.functions.append(func)
        return func

    def define_function_alias(self, var_name: str, func_name: str) -> None:
        if any(alias.alias == var_name for alias in self.function_aliasses):
            return None
        self.function_aliasses.append(FunctionAlias(var_name, func_name))

    def get_function_alias(self, var_name: str) -> Optional[str]:
        alias = next((alias for alias in self.function_aliasses if alias.alias == var_name), None)
        if alias is None and self.parent is not None:
            alias = self.parent.get_function_alias(var_name)
        return alias

    def get_function_by_alias(self, alias: str) -> Optional[CTFunction]:
        alias = self.get_function_alias(alias)
        if alias is None:
            return None
        return self.get_function(alias.name)

    # need to give this another name if we keep the latter.
    def get_function(self, name: str) -> Optional[CTFunction]:
        func = next((func for func in self.functions if func.name == name), None)
        if func is None and self.parent is not None:
            func = self.parent.get_function(name)
        return func

    def get_function_by_address(self, address: int) -> Optional[CTFunction]:
        func = next((func for func in self.functions if func.start_address == address), None)
        if func is None and self.parent is not None:
            func = self.parent.get_function_by_address(address)
        return func

    def define_library(self, name: str) -> Optional[int]:
        if name in self.root._libraries:
            return None
        id = len(self.root._libraries)
        self.root._libraries[name] = id
        return id

    def get_library(self, name: str) -> Optional[int]:
        if name in self.root._libraries:
            return self.root._libraries[name]
        return None

    def __str__(self) -> str:
        return f"CompilerScope({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def discard(self) -> None:
        if self.parent != None:
            self.parent.remove_child(self)